import numpy as np
import math
import warnings
from scipy.optimize import curve_fit
from collections import defaultdict
from opynfield.config.user_input import UserInput
from opynfield.config.defaults_settings import Defaults
from opynfield.calculate_measures import coverage_math
from opynfield.config.cov_asymptote import CoverageAsymptote
from opynfield.readin.track import Track
from opynfield.calculate_measures.standard_track import StandardTrack


def cartesian_to_polar(
    x: np.ndarray, y: np.ndarray, verbose: bool
) -> tuple[np.ndarray, np.ndarray]:
    """ This function converts an animal's tracking coordinates from cartesian to polar coordinates.
    It is one step in the track standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param x: the x coordinates of a :class:`opynfield.readin.track.Track` object
    :type x: np.ndarray
    :param y: the y coordinates of a :class:`opynfield.readin.track.Track` object
    :type y: np.ndarray
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: a tuple of the r and theta coordinates of the :class:`opynfield.readin.track.Track`
    :rtype: tuple[np.ndarray, np.ndarray]
    """

    # initialize results
    radius = np.zeros(shape=x.shape)
    angle = np.zeros(shape=x.shape)
    # for each point
    for i in range(len(x)):
        # find distance from (0, 0)
        rad = np.sqrt(x[i] ** 2 + y[i] ** 2)
        # find angle from positive x-axis
        theta_radians = np.arctan2(y[i], x[i])
        theta_degrees = math.degrees(theta_radians)
        radius[i] = rad
        angle[i] = theta_degrees
    if verbose:
        print("Polar Coordinates Calculated")
    return radius, angle


def step_distance(x: np.ndarray, y: np.ndarray, verbose: bool) -> np.ndarray:
    """This function calculates the distance between each consecutive pair of tracking coordinates of an animal.
    It is one step in the track standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param x: the x coordinates of a :class:`opynfield.readin.track.Track` object
    :type x: np.ndarray
    :param y: the y coordinates of a :class:`opynfield.readin.track.Track` object
    :type y: np.ndarray
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: the distance stepped between tracking points (also called 'activity')
    :rtype: np.ndarray
    """
    # initialize results
    act = np.zeros(shape=x.shape)
    # for each point
    for i in range(len(x) - 1):
        # distance is defined between two points, so we will have a trailing zero
        # find the distance between it and the next point
        dx = x[i + 1] - x[i]
        dy = y[i + 1] - y[i]
        dist = np.sqrt(dx**2 + dy**2)
        act[i] = dist
    if verbose:
        print("Activity Calculated")
    return act


def turning_angle(x: np.ndarray, y: np.ndarray, verbose: bool) -> np.ndarray:
    """This function calculates the angle that the animal turned between two consecutive steps,
    using the law of cosines. It is one step in the track standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param x: the x coordinates of a :class:`opynfield.readin.track.Track` object
    :type x: np.ndarray
    :param y: the y coordinates of a :class:`opynfield.readin.track.Track` object
    :type y: np.ndarray
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: the magnitude of the angle (in degrees) from a straight heading that the animal turned between
        each consecutive set of steps.
    :rtype: np.ndarray
    """
    # initialize results
    turn_ang = np.zeros(shape=x.shape)
    # for each point
    for i in range(len(x) - 2):
        # a turn is defined between two steps or 3 points, so we will have a leading and a trailing zero
        x1 = x[i]
        x2 = x[i + 1]
        x3 = x[i + 2]
        y1 = y[i]
        y2 = y[i + 1]
        y3 = y[i + 2]
        # use law of cosines to get the angle based on the side lengths
        # A is the distance (triangle side length) between previous and current pts
        a = np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        # B is the distance (triangle side length) between current and next pts
        b = np.sqrt((x3 - x2) ** 2 + (y3 - y2) ** 2)
        # C is the distance (triangle side length) between previous and next pts
        c = np.sqrt((x3 - x1) ** 2 + (y3 - y1) ** 2)
        # law of cosines: c^2 = a^2 + b^2 - 2abcos(angleC)
        # --> angleC = arccos((a^2 + b^2 - c^2)/(2ab))
        # t1 is numerator
        t1 = (a**2) + (b**2) - (c**2)
        # t2 is denominator
        t2 = a * b * 2
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            # if the fly doesn't move during one of the steps, the triangle is
            # undefined this allows the div by zero to not send warning message
            # bc the arccos (inf) will be nan anyway
            b_mat = t1 / t2
            # if the angle is out of range we want nan to be there instead
            # this silences the warning message about nan
            ta = np.arccos(b_mat)
        # converts from angle C to turn angle
        ta_deg = math.degrees(abs(np.pi - ta))
        # we only care about the magnitude of the turn
        turn_ang[i + 1] = ta_deg  # save to the point where the turn was made
    if verbose:
        print("Turning Angles Calculated")
    return turn_ang


def motion_probabilities(
    radius: np.ndarray,
    act: np.ndarray,
    turn_angle: np.ndarray,
    inactivity_threshold: float,
    edge_rad: float,
    verbose: bool,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """This function calculates P++, P+-, P+0, P0+, and P00 (collectively 'motion probabilities')
    from an animal's track trajectory. See motion probability types for more information. It is one step in the track
    standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param radius: the radial position of the animal at each tracking point
    :type radius: np.ndarray
    :param act: the length of the step taken by the animal between two points
    :type act: np.ndarray
    :param turn_angle: the angle the animal turned at each point between two steps
    :type turn_angle: np.ndarray
    :param inactivity_threshold: to account for body wobble, how small of a step length do we consider to be
        negligible and set to zero - sourced from :class:`opynfield.config.user_input.UserInput` object
    :type inactivity_threshold: np.ndarray
    :param edge_rad: to consider the motion probabilities only near the edge of the arena, at what radius do we consider
        the animal to be in the edge region - sourced from :class:`opynfield.config.user_input.UserInput` object
    :type edge_rad: np.ndarray
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: tuple of the raw data for motion probabilities (P++, P+-, P+0, P0+, P00) - 1 where the action occurred and
        0 elsewhere
    :rtype: tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
    """
    # initialize results
    ppp = np.zeros(shape=radius.shape)
    ppm = np.zeros(shape=radius.shape)
    ppz = np.zeros(shape=radius.shape)
    pzp = np.zeros(shape=radius.shape)
    pzz = np.zeros(shape=radius.shape)

    # all zeros, fill in ones where condition is true
    for i in range(1, len(radius) - 1):
        # a motion probability is defined between two steps or 3 points, so we will have a leading and a trailing zero
        if radius[i] >= edge_rad:
            # the turn decision is made inside the edge region
            # (all decisions in the center of the arena do not count towards the motion probabilities)
            step_1 = act[i - 1]
            step_2 = act[i]
            if step_1 > inactivity_threshold:
                # will be p plus something
                if step_2 > inactivity_threshold:
                    # will be p plus plus or p plus minus
                    if turn_angle[i] < 90:
                        # considered continuing "straight", so this is p++
                        ppp[i] = 1
                    else:
                        # they turned around
                        ppm[i] = 1
                else:
                    # this is p+0
                    ppz[i] = 1
            else:
                # will be p zero something
                if step_2 > inactivity_threshold:
                    # this is p0+
                    pzp[i] = 1
                else:
                    # this is p00
                    pzz[i] = 1
    if verbose:
        print("Motion Probabilities Calculated")
    return ppp, ppm, ppz, pzp, pzz


def locate_bin(
    radius: np.ndarray,
    theta_angle: np.ndarray,
    node_size: float,
    edge_rad: float,
    verbose: bool,
) -> tuple[np.ndarray, float]:
    """This function assigns a bin number to each tracking point of an animal to facilitate the calculation of coverage.
    It is one step in the track standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param radius: the radial position of the animal at each tracking point
    :type radius: np.ndarray
    :param theta_angle: the angular position of the animal at each tracking point
    :type theta_angle: np.ndarray
    :param node_size: the angle that (when combined with ``edge_rad``) defines a bin - sourced from
        :class:`opynfield.config.user_input.UserInput`
    :type node_size: float
    :param edge_rad: at what radius do we consider the animal to be in the edge region, when combined with ``node_size``
        this defines a bin - sourced from :class:`opynfield.config.user_input.UserInput`
    :type edge_rad: float
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: tuple of the bin assignment array and the total number of bins used to make the assignment
    :rtype: tuple[np.ndarray, float]
    """
    # initialize bins as -1 (an invalid bin number)
    bins = np.ones(shape=radius.shape) * -1

    num_bins = 360 / node_size  # number of bins the arena edge is divided into
    if num_bins < 8:
        raise ValueError("Node Size Too Small")
    for i in range(len(radius)):
        if radius[i] >= edge_rad:
            # the point is in the edge region, so we need to assign it a bin
            bin_i = np.floor(theta_angle[i] / node_size)
            bins[i] = bin_i
    if verbose:
        print("Coverage Bins Located")
    return bins, num_bins


def calculate_coverage(
    cov_bins: np.ndarray, num_bins: float, verbose: bool
) -> np.ndarray:
    """This function is calls functions from the coverage_math module to calculate the animal's coverage over time.
    See coverage types for comparison of (raw) coverage to other coverage measures.
    It is one step in the track standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param cov_bins: which bin the animal is located in at each tracking point
    :type cov_bins: np.ndarray
    :param num_bins: the total number of bins the arena edge region is divided into
    :type num_bins: float
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: the coverage of the animal at each tracking point
    :rtype: np.ndarray
    """
    # locate_bin has already worried about if the animal is in edge or not, this is just looking at visits and cov
    cov = coverage_math.path_coverage(cov_bins, num_bins)
    if verbose:
        print("Coverage Calculated")
    return cov


def calculate_percent_coverage(cov: np.ndarray, verbose: bool) -> np.ndarray:
    """This function normalizes an animal's coverage by the maximum coverage the animal reaches to get its percent
    coverage. See coverage types for comparison of percent coverage to other coverage measures.
    It is one step in the track standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param cov: the coverage of the animal at each tracking point
    :type cov: np.ndarray
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: the percent coverage (normalized coverage) of the animal at each tracking point
    :rtype: np.ndarray
    """
    highest_cov = max(cov)
    perc_cov = cov / highest_cov
    if verbose:
        print("Percent Coverage Calculated")
    return perc_cov


def calculate_pica(
    cov: np.ndarray, asymptote_info: CoverageAsymptote, verbose: bool
) -> tuple[np.ndarray, float]:
    """This function normalizes an animal's coverage by its predicted coverage asymptote from the time vs coverage
    model. See coverage types for comparison of PICA to other coverage measures.
    It is one step in the track standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param cov: the coverage of the animal at each tracking point
    :type cov: np.ndarray
    :param asymptote_info: a custom dataclass that includes information needed to calculate the animal's
        coverage asymptote
    :type asymptote_info: CoverageAsymptote
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: the PICA (Percent of Individual Coverage Asymptote) of the animal at each tracking point
    :rtype: np.ndarray
    """
    # format the input data to the fit
    x2 = np.arange(len(cov))
    x1 = np.array(x2)
    y1 = np.array(cov)
    y = y1[~np.isnan(y1)]
    x = x1[~np.isnan(y1)]
    # fit a model to the coverage data
    # noinspection PyTupleAssignmentBalance
    params, cv = curve_fit(
        asymptote_info.f_name,
        x,
        y,
        p0=asymptote_info.initial_parameters,
        bounds=asymptote_info.parameter_bounds,
        **{"maxfev": asymptote_info.max_f_eval}
    )
    # extract the asymptote value from the model
    asymptote_i = params[asymptote_info.asymptote_param] * asymptote_info.asymptote_sign
    # calculate pica
    pica_i = cov / asymptote_i
    if verbose:
        print("PICA Calculated")
    return pica_i, asymptote_i


def calculate_group_coverage_asymptote(
    group_tracks: list[StandardTrack], asymptote_info: CoverageAsymptote
) -> float:
    """This function calculates the coverage asymptote of a group based on the aggregated data from all
    individuals in the group.

    :param group_tracks: a list of all standardized tracks belonging to the same analysis group
    :type group_tracks: list[StandardTrack]
    :param asymptote_info: a custom dataclass that includes information needed to calculate the group's
        coverage asymptote
    :type asymptote_info: CoverageAsymptote
    :return: the group's coverage asymptote to be used in calculating the PGCA of individuals in that group
    :rtype: float
    """
    group_coverage = list()
    group_time = list()
    for track_g in group_tracks:
        for i in range(len(track_g.coverage)):
            group_coverage.append(track_g.coverage[i])
            group_time.append(track_g.t[i])
    x1 = np.array(group_time)
    y1 = np.array(group_coverage)
    y = y1[~np.isnan(y1)]
    x = x1[~np.isnan(y1)]
    # noinspection PyTupleAssignmentBalance
    params, cv = curve_fit(
        asymptote_info.f_name,
        x,
        y,
        asymptote_info.initial_parameters,
        bounds=asymptote_info.parameter_bounds,
        **{"maxfev": asymptote_info.max_f_eval}
    )
    # extract the asymptote value from the model
    asymptote_g = params[asymptote_info.asymptote_param] * asymptote_info.asymptote_sign
    return asymptote_g


def calculate_pgca(s_track: StandardTrack, asymptote_g: float, verbose: bool):
    """This function normalizes an animal's coverage by its group's predicted coverage asymptote from the time vs
    coverage model. It also sets the group asymptote as an attribute of the track.
    This must occur after all tracks are standardized and group coverage asymptotes are calculated.
    See coverage types for comparison of PGCA to other coverage measures.

    :param s_track: a standardized track
    :type s_track: StandardTrack
    :param asymptote_g: the group coverage asymptote for the group to which the s_track belongs
        coverage asymptote
    :type asymptote_g: float
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: the PGCA (Percent of Group Coverage Asymptote) of the animal at each tracking point
    :rtype: np.ndarray
    """
    track_pgca = s_track.coverage / asymptote_g
    s_track.set_pgca(track_pgca, asymptote_g)
    if verbose:
        print("PGCA Calculated")
    pass


def motion_probabilities_given_previous(
    ppp: np.ndarray, ppm: np.ndarray, ppz: np.ndarray, pzp: np.ndarray, pzz: np.ndarray
) -> tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray
]:
    """This function calculates P++Given+, P+-Given+, P+0Given+, P0+Given0, P00Given0, P++GivenAny, P+-GivenAny,
    P+0GivenAny, P0+GivenAny, and P00GivenAny (collectively 'motion probabilities given previous'). See motion
    probability types for more information. It is one step in the track standardization procedure orchestrated
    by :func:`opynfield.calculate_measures.calculate_measures.tracks_to_measures`

    :param ppp: the raw P++ values from :func:`opynfield.calculate_measures.calculate_measures.motion_probabilities`
    :type ppp: np.ndarray
    :param ppm: the raw P+- values from :func:`opynfield.calculate_measures.calculate_measures.motion_probabilities`
    :type ppm: np.ndarray
    :param ppz: the raw P+0 values from :func:`opynfield.calculate_measures.calculate_measures.motion_probabilities`
    :type ppz: np.ndarray
    :param pzp: the raw P0+ values from :func:`opynfield.calculate_measures.calculate_measures.motion_probabilities`
    :type pzp: np.ndarray
    :param pzz: the raw P00 values from :func:`opynfield.calculate_measures.calculate_measures.motion_probabilities`
    :type pzz: np.ndarray
    :return: the motion probabilities given previous - 1 where the action occurred, 0 where it
        could have occurred but didn't, and np.nan elsewhere.
    :rtype: tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray,
        np.ndarray, np.ndarray]
    """
    # take the 'raw' motion probs (zero everywhere but one where mp occurred)
    # convert to given plus, given zero, and given any formats
    # np.nans everywhere, 1 where mp occurred, 0 where mp could have occurred

    # previous steps
    given_plus = ppp + ppm + ppz
    given_zero = pzp + pzz
    given_any = given_plus + given_zero

    # ppp given plus should first 0 where given_plus is true and np.nan where it is not
    # then it should be 1 where ppp is true and 0/nan (ppp given plus) where not
    pppgp = np.where(given_plus == 1, 0, np.nan)
    pppgp = np.where(ppp == 1, 1, pppgp)
    # ppm given plus should first 0 where given_plus is true and np.nan where it is not
    # then it should be 1 where ppm is true and 0/nan (ppm given plus) where not
    ppmgp = np.where(given_plus == 1, 0, np.nan)
    ppmgp = np.where(ppm == 1, 1, ppmgp)
    # ppz given plus should first 0 where given_plus is true and np.nan where it is not
    # then it should be 1 where ppz is true and 0/nan (ppz given plus) where not
    ppzgp = np.where(given_plus == 1, 0, np.nan)
    ppzgp = np.where(ppz == 1, 1, ppzgp)
    # pzp given zero should first 0 where given_zero is true and np.nan where it is not
    # then it should be 1 where pzp is true and 0/nan (pzp given zero) where not
    pzpgz = np.where(given_zero == 1, 0, np.nan)
    pzpgz = np.where(pzp == 1, 1, pzpgz)
    # pzz given zero should first 0 where given_zero is true and np.nan where it is not
    # then it should be 1 where pzz is true and 0/nan (pzz given zero) where not
    pzzgz = np.where(given_zero == 1, 0, np.nan)
    pzzgz = np.where(pzz == 1, 1, pzzgz)

    # ppp given any should first 0 where given_any is true and np.nan where it is not
    # then it should be 1 where ppp is true and 0/nan (ppp given any) where not
    pppga = np.where(given_any == 1, 0, np.nan)
    pppga = np.where(ppp == 1, 1, pppga)
    # ppm given any should first 0 where given_any is true and np.nan where it is not
    # then it should be 1 where ppm is true and 0/nan (ppm given any) where not
    ppmga = np.where(given_any == 1, 0, np.nan)
    ppmga = np.where(ppm == 1, 1, ppmga)
    # ppz given any should first 0 where given_any is true and np.nan where it is not
    # then it should be 1 where ppz is true and 0/nan (ppz given any) where not
    ppzga = np.where(given_any == 1, 0, np.nan)
    ppzga = np.where(ppz == 1, 1, ppzga)
    # pzp given any should first 0 where given_any is true and np.nan where it is not
    # then it should be 1 where pzp is true and 0/nan (pzp given any) where not
    pzpga = np.where(given_any == 1, 0, np.nan)
    pzpga = np.where(pzp == 1, 1, pzpga)
    # pzz given any should first 0 where given_any is true and np.nan where it is not
    # then it should be 1 where pzz is true and 0/nan (pzz given any) where not
    pzzga = np.where(given_any == 1, 0, np.nan)
    pzzga = np.where(pzz == 1, 1, pzzga)
    return pppgp, ppmgp, ppzgp, pzpgz, pzzgz, pppga, ppmga, ppzga, pzpga, pzzga


def tracks_to_measures(
    all_tracks: list[Track],
    user_config: UserInput,
    default_settings: Defaults,
    coverage_settings: CoverageAsymptote,
) -> tuple[list[StandardTrack], defaultdict[str, list[StandardTrack]]]:
    """This function runs the entire standardization procedure to transform a Track to a StandardTrack. It runs the
    procedure for a list of all tracks in the analysis.

    :param all_tracks: list of all the tracks read in for analysis
    :type all_tracks: list[Track]
    :param user_config: a custom dataclass that contains user inputs for analysis
    :type user_config: UserInput
    :param default_settings: a custom dataclass that contains default settings for analysis
    :type default_settings: Defaults
    :param coverage_settings: a custom class that contains settings for the calculation of coverage asymptotes
    :type coverage_settings: CoverageAsymptote
    :return: tuple of  a list of all standardized tracks and a dictionary of group names to a list of all standardized
        tracks belonging to that group
    :rtype: tuple[list[StandardTrack], defaultdict[str, list[StandardTrack]]]
    """
    tracks_by_group = defaultdict(list)
    all_standard_tracks = list()
    for track in all_tracks:
        assert track.standardized
        # calculate all the independent measures for the track
        r, theta = cartesian_to_polar(track.x, track.y, user_config.verbose)
        activity = step_distance(track.x, track.y, user_config.verbose)
        turn = turning_angle(track.x, track.y, user_config.verbose)
        (
            p_plus_plus,
            p_plus_minus,
            p_plus_zero,
            p_zero_plus,
            p_zero_zero,
        ) = motion_probabilities(
            r,
            activity,
            turn,
            user_config.inactivity_threshold,
            user_config.set_edge_radius(),
            user_config.verbose,
        )
        (
            p_plus_plus_given_plus,
            p_plus_minus_given_plus,
            p_plus_zero_given_plus,
            p_zero_plus_given_zero,
            p_zero_zero_given_zero,
            p_plus_plus_given_any,
            p_plus_minus_given_any,
            p_plus_zero_given_any,
            p_zero_plus_given_any,
            p_zero_zero_given_any,
        ) = motion_probabilities_given_previous(
            p_plus_plus, p_plus_minus, p_plus_zero, p_zero_plus, p_zero_zero
        )
        coverage_bins, n_bins = locate_bin(
            r,
            theta,
            default_settings.node_size,
            user_config.set_edge_radius(),
            user_config.verbose,
        )
        coverage = calculate_coverage(coverage_bins, n_bins, user_config.verbose)
        percent_coverage = calculate_percent_coverage(coverage, user_config.verbose)
        pica, asymptote = calculate_pica(
            coverage, coverage_settings, user_config.verbose
        )
        # save to standard track with all measures as attributes
        # initialize the pgca as ones and np.nan until they are calculated
        standard_track = StandardTrack(
            track.group,
            track.x,
            track.y,
            track.t,
            r,
            theta,
            activity,
            turn,
            p_plus_plus,
            p_plus_minus,
            p_plus_zero,
            p_zero_plus,
            p_zero_zero,
            coverage_bins,
            n_bins,
            coverage,
            percent_coverage,
            pica,
            asymptote,
            np.ones(shape=pica.shape),
            np.nan,
            p_plus_plus_given_plus,
            p_plus_minus_given_plus,
            p_plus_zero_given_plus,
            p_zero_plus_given_zero,
            p_zero_zero_given_zero,
            p_plus_plus_given_any,
            p_plus_minus_given_any,
            p_plus_zero_given_any,
            p_zero_plus_given_any,
            p_zero_zero_given_any,
        )
        all_standard_tracks.append(standard_track)
        # create list of tracks in that group
        tracks_by_group[standard_track.group].append(standard_track)
    if user_config.verbose:
        print("All Independent Track Measures Calculated")
        print("Tracks Converted To Standard Tracks")
    # now find the coverage asymptote for the whole group to get pgca
    for g in tracks_by_group:
        # this group's asymptote
        group_asymptote = calculate_group_coverage_asymptote(
            tracks_by_group[g], coverage_settings
        )
        for track in tracks_by_group[g]:
            # use the group's asymptote to calculate pgca for each track in that group
            calculate_pgca(track, group_asymptote, user_config.verbose)
    return all_standard_tracks, tracks_by_group



import numpy as np
import math
import pandas as pd
from scipy import interpolate
import warnings


def running_line(y_array: np.ndarray, n: int, dn: int) -> np.ndarray:
    """This function smooths the tracking data in the same way that ethovision automatically smooths tracking data in
    order to minimize the impact of body wobble

    :param y_array: the input coordinates
    :type y_array: np.ndarray
    :param n: the running window length
    :type n: int
    :param dn: the window step size
    :type dn: int
    :return: the smoothed coordinates
    :rtype: np.ndarray
    """
    # based on the MATLAB runline function
    # y is the input coordinate
    # n is the length of the running window in samples
    # dn is the step size of the window in samples

    num_points = len(y_array)  # number of points in the y_array
    y_line = np.zeros(shape=num_points)  # initialized results
    norm = np.zeros(shape=num_points)  # ??
    num_windows = (
        math.ceil((num_points - n) / dn) + 1
    )  # number of windows to smooth within
    y_fit = np.zeros(shape=(num_windows, n))  # the fit for a particular window

    # define weights with tri-weight distribution
    points = np.arange(0, n, 1)
    h = n / 2
    mid_point = points.mean()
    x_weight_dist = (points - mid_point) / h

    # Tukey tri-weight kernel
    y_weights = (1 - (abs(x_weight_dist) ** 3)) ** 3

    # for each window
    for window in range(num_windows):
        y_segment = y_array[
            window : window + 5
        ]  # take the segment of the data from that window
        y1 = y_segment.mean()
        y2 = (np.arange(1, n + 1) * y_segment).mean() * 2 / (n + 1)
        m_model = (y2 - y1) * 6 / (n - 1)
        b_model = y1 - m_model * (n + 1) / 2
        y_fit[window, :] = (np.arange(1, n + 1) * m_model) + b_model
        y_line[window : window + n] = y_line[window : window + n] + (
            y_fit[window, :] * y_weights
        )
        norm[window : window + n] = norm[window : window + n] + y_weights

    # for the remaining points that didn't fit in a full window
    mask = np.nonzero(norm > 0)
    y_line[mask] = y_line[mask] / norm[mask]
    index = (num_windows - 1) * dn + (n - 1)
    num_end_pts = len(y_array) - index + 1
    # noinspection PyUnboundLocalVariable
    y_line[np.arange(index - 1, num_points)] = (
        np.arange(n + 1, n + num_end_pts + 1) * m_model + b_model
    )

    return y_line


def subsample(coord, sample_freq: int, sample_interval: int) -> np.ndarray:
    """This function sub-samples the recorded coordinates to the desired data density

    :param coord: the full-density coordinates
    :type coord: np.ndarray
    :param sample_freq: the sampling frequency with which the data was recorded
    :type sample_freq: int
    :param sample_interval: the desired density
    :type sample_interval: int
    :return: the sub-sampled coordinates
    :rtype: np.ndarray
    """
    # sample freq is in Hz (samples per second)
    # sample_interval is time (number of seconds you want between points)
    num_pts_per_int = int(sample_interval / (1 / sample_freq))
    # how many points at given sampling frequency are between points at desired sampling interval
    coord_subsample = coord[0::num_pts_per_int].copy()
    # the subsampled coordinate
    return coord_subsample


def fill_missing_data(coord: np.ndarray, time: np.ndarray) -> np.ndarray:
    """This function identified missing data points and interpolated the animals position at that point

    :param coord: the coordinates
    :type coord: np.ndarray
    :param time: the time coordinate
    :type time: np.ndarray
    :return: the interpolated coordinates
    :rtype: np.ndarray
    """
    # takes in either the x or y coordinate and time
    # interpolates missing values where the fly was not tracked
    time_orig = time
    # delete x where y is nan
    time = time[~np.isnan(coord)]
    # delete x where x is nan
    time = time[~np.isnan(time)]
    # delete y where y is nan
    coord = coord[~np.isnan(coord)]
    # fill in interpolated values
    f = interpolate.interp1d(time, coord, fill_value="extrapolate")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        # catch warnings bc if x had nan y could not be calculated
        coord_new = f(time_orig)
    coord_new = pd.Series(coord_new)
    return coord_new.values


def calc_center(
    combined_x: np.ndarray, combined_y: np.ndarray, verbose: bool, trim: int = 0
) -> tuple[float, float]:
    """This function estimates the center point of an arena from tracking coordinates

    :param combined_x: the x coordinates to estimate from
    :param combined_y: the y coordinates the estimate from
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :param trim: how many points are recorded before the animal enters the arena (maximum of all animals in this track type)
    :type trim: int, defaults to 0
    :return: the center point of the arena
    :rtype: tuple[float, float]
    """
    # this is a rough calculation of the center point
    # would be better to calculate a minimum enclosing circle and compare rough to precise calc
    x_cen_rough = np.nanmin(combined_x[trim:]) + (
        (np.nanmax(combined_x[trim:]) - np.nanmin(combined_x[trim:])) / 2
    )
    y_cen_rough = np.nanmin(combined_y[trim:]) + (
        (np.nanmax(combined_y[trim:]) - np.nanmin(combined_y[trim:])) / 2
    )
    if verbose:
        print("Combined Coordinate Center Point Calculated")
    return x_cen_rough, y_cen_rough



from dataclasses import dataclass, field

import numpy as np
import pandas as pd

from opynfield.readin import multi_tracker


@dataclass
class Track:
    """This dataclass aggregates information from a single track once the coordinates are standardized, but before other
    measures are calculated

    Attributes:
        group (str): the group to which the track belongs
        x (np.ndarray): the x coordinates of the track
        y (np.ndarray): the y coordinates of the track
        t (np.ndarray): the time coordinates of the track
        track_type (str): the recording type of the track
        options (list): additional information depending on the track type
        standardized (bool): has the track finished its standardization process?

    Methods:
        buri_convert_units
        buri_convert_to_center
        buri_running_line
        buri_subsample
        buri_fill_missing
        etho_v1_numeric
        etho_v1_subsample
        etho_v1_fill_missing
        etho_v1_convert_to_center
        etho_v2_numeric
        etho_v2_subsample
        etho_v2_fill_missing
        etho_v2_convert_to_center
        etho_txt_numeric
        etho_txt_subsample
        etho_txt_fill_missing
        etho_txt_convert_to_center
        etho_ml_numeric
        etho_ml_subsample
        etho_ml_fill_missing
        etho_ml_convert_to_center
        anymaze_center_numeric
        anymaze_center_running_line
        anymaze_center_subsample
        anymaze_center_fill_missing
        anymaze_center_convert_to_center
        anymaze_center_convert_units
        anymaze_head_numeric
        anymaze_head_running_line
        anymaze_head_subsample
        anymaze_head_fill_missing
        anymaze_head_convert_to_center
        anymaze_head_convert_units
    """
    # dataclass for standardized track after data is read in and wrangled
    group: str  # can change to enum
    x: np.ndarray = field(repr=False)  # x coordinate
    y: np.ndarray = field(repr=False)  # y coordinate
    t: np.ndarray = field(repr=False)  # time
    track_type: str  # which data type this came from -> can change to enum
    options: list
    # for buri: x and y center coordinates (in cm) then arena_radius_px
    # for ethov1: arena number
    # for ethov2: arena number
    standardized: bool  # has the track been completely standardized yet

    def buri_convert_units(
        self, arena_radius_cm: float, arena_radius_px: int, verbose: bool
    ):
        """Converts the units of the track"""
        assert self.track_type == "Buridian Tracker"
        self.x = self.x * (
            arena_radius_cm / arena_radius_px
        )  # convert x coordinate from pixels to cm
        self.y = self.y * (
            arena_radius_cm / arena_radius_px
        )  # convert y coordinate from pixels to cm
        self.t = self.t / 1000  # convert time from ms to s
        if verbose:
            print("Buri Units Converted")

    def buri_convert_to_center(
        self, center_point_x: float, center_point_y: float, verbose: bool
    ):
        """Centers the coordinates based on the arena center point"""
        assert self.track_type == "Buridian Tracker"
        # subtract the center point so that the track coordinate system will be centered at (0,0)
        self.x = self.x - center_point_x
        self.y = self.y - center_point_y
        if verbose:
            print("Buri Units Centered")

    def buri_running_line(
        self, running_window_length: int, window_step_size: int, verbose: bool
    ):
        """Smooths the coordinates using the running line formula"""
        assert self.track_type == "Buridian Tracker"
        # smooth the coordinates using the same function used in ethovision tracks
        self.x = multi_tracker.running_line(
            self.x, running_window_length, window_step_size
        )
        self.y = multi_tracker.running_line(
            self.y, running_window_length, window_step_size
        )
        if verbose:
            print("Buri Track Smoothed")

    def buri_subsample(self, sample_freq: int, time_bin_size: int, verbose: bool):
        """Sub-samples the coordinates to the desired density"""
        assert self.track_type == "Buridian Tracker"
        # subsample the coordinates to the desired sampling frequency
        self.x = multi_tracker.subsample(self.x, sample_freq, time_bin_size)
        self.y = multi_tracker.subsample(self.y, sample_freq, time_bin_size)
        self.t = multi_tracker.subsample(self.t, sample_freq, time_bin_size)
        if verbose:
            print("Buri Track Subsampled")

    def buri_fill_missing(self, verbose: bool):
        """Fills in the missing coordinates"""
        assert self.track_type == "Buridian Tracker"
        # fill missing data using linear extrapolation
        self.x = multi_tracker.fill_missing_data(self.x, self.t)
        self.y = multi_tracker.fill_missing_data(self.y, self.t)
        if verbose:
            print("Buri Track Missing Values Filled")

    def etho_v1_numeric(self, verbose: bool):
        """Changes datatype to numeric"""
        assert self.track_type == "Ethovision Excel Version 1"
        # in v1 the values are stored as floats but there are string '-' for missing values that we set to nan
        for i in range(len(self.x)):
            if type(self.x[i]) != float:
                self.x[i] = np.nan
            if type(self.y[i]) != float:
                self.y[i] = np.nan
        self.x = self.x.astype(np.float64)
        self.y = self.y.astype(np.float64)
        self.t = self.t.astype(np.float64)
        if verbose:
            print("Etho V1 Track Converted To Numeric")

    def etho_v1_subsample(self, sample_freq: int, time_bin_size: int, verbose: bool):
        """Sub-samples the coordinates to the desired density"""
        assert self.track_type == "Ethovision Excel Version 1"
        self.x = multi_tracker.subsample(self.x, sample_freq, time_bin_size)
        self.y = multi_tracker.subsample(self.y, sample_freq, time_bin_size)
        self.t = multi_tracker.subsample(self.t, sample_freq, time_bin_size)
        if verbose:
            print("Etho V1 Track Subsampled")

    def etho_v1_fill_missing(self, verbose: bool):
        """Fills in the missing coordinates"""
        assert self.track_type == "Ethovision Excel Version 1"
        self.x = multi_tracker.fill_missing_data(self.x, self.t)
        self.y = multi_tracker.fill_missing_data(self.y, self.t)
        if verbose:
            print("Etho V1 Track Missing Values Filled")

    def etho_v1_convert_to_center(
        self, center_points_by_area: dict[str, tuple[float, float]], verbose: bool
    ):
        """Centers the coordinates based on the arena center point"""
        assert self.track_type == "Ethovision Excel Version 1"
        self.x = (
            self.x - center_points_by_area[self.options[0]][0]
        )  # x coordinate of the arena's center point
        self.y = (
            self.y - center_points_by_area[self.options[0]][1]
        )  # x coordinate of the arena's center point
        if verbose:
            print("Etho V1 Track Centered")

    def etho_v2_numeric(self, verbose):
        """Changes datatype to numeric"""
        # in v1 the values are stored as number strings, so we try to convert to float
        # but there are string '-' for missing values that won't convert to float that we set to nan
        for i in range(len(self.x)):
            try:
                self.x[i] = float(self.x[i])
            except ValueError:
                self.x[i] = np.nan
            try:
                self.y[i] = float(self.y[i])
            except ValueError:
                self.y[i] = np.nan
        self.x = self.x.astype(np.float64)
        self.y = self.y.astype(np.float64)
        self.t = self.t.astype(np.float64)
        if verbose:
            print("Etho V2 Track Converted To Numeric")

    def etho_v2_subsample(self, sample_freq: int, time_bin_size: int, verbose: bool):
        """Sub-samples the coordinates to the desired density"""
        assert self.track_type == "Ethovision Excel Version 2"
        self.x = multi_tracker.subsample(self.x, sample_freq, time_bin_size)
        self.y = multi_tracker.subsample(self.y, sample_freq, time_bin_size)
        self.t = multi_tracker.subsample(self.t, sample_freq, time_bin_size)
        if verbose:
            print("Etho V2 Track Subsampled")

    def etho_v2_fill_missing(self, verbose: bool):
        """Fills in the missing coordinates"""
        assert self.track_type == "Ethovision Excel Version 2"
        self.x = multi_tracker.fill_missing_data(self.x, self.t)
        self.y = multi_tracker.fill_missing_data(self.y, self.t)
        if verbose:
            print("Etho V2 Track Missing Values Filled")

    def etho_v2_convert_to_center(
        self, center_points_by_area: dict[str, tuple[float, float]], verbose: bool
    ):
        """Centers the coordinates based on the arena center point"""
        assert self.track_type == "Ethovision Excel Version 2"
        self.x = (
            self.x - center_points_by_area[self.options][0]
        )  # x coordinate of the arena's center point
        self.y = (
            self.y - center_points_by_area[self.options][1]
        )  # x coordinate of the arena's center point
        if verbose:
            print("Etho V2 Track Centered")

    def etho_txt_numeric(self, verbose: bool):
        """Changes datatype to numeric"""
        self.x = pd.to_numeric(self.x, errors="coerce")
        self.y = pd.to_numeric(self.y, errors="coerce")
        self.t = pd.to_numeric(self.t, errors="coerce")
        if verbose:
            print("Etho Text Track Converted To Numeric")

    def etho_txt_subsample(self, sample_freq: int, time_bin_size: int, verbose: bool):
        """Sub-samples the coordinates to the desired density"""
        assert self.track_type == "Ethovision Text"
        self.x = multi_tracker.subsample(self.x, sample_freq, time_bin_size)
        self.y = multi_tracker.subsample(self.y, sample_freq, time_bin_size)
        self.t = multi_tracker.subsample(self.t, sample_freq, time_bin_size)
        if verbose:
            print("Etho Text Track Subsampled")

    def etho_txt_fill_missing(self, verbose: bool):
        """Fills in the missing coordinates"""
        assert self.track_type == "Ethovision Text"
        self.x = multi_tracker.fill_missing_data(self.x, self.t)
        self.y = multi_tracker.fill_missing_data(self.y, self.t)
        if verbose:
            print("Etho Text Track Missing Values Filled")

    def etho_txt_convert_to_center(
        self, center_points_by_area: dict[str, tuple[float, float]], verbose: bool
    ):
        """Centers the coordinates based on the arena center point"""
        assert self.track_type == "Ethovision Text"
        self.x = (
            self.x - center_points_by_area[self.options][0]
        )  # x coordinate of the arena's center point
        self.y = (
            self.y - center_points_by_area[self.options][1]
        )  # x coordinate of the arena's center point
        if verbose:
            print("Etho Text Track Centered")

    def etho_ml_numeric(self, verbose: bool):
        """Changes datatype to numeric"""
        assert self.track_type == "Ethovision Through MATLAB"
        for i in range(len(self.x)):
            if type(self.x[i]) != np.float64:
                self.x[i] = np.nan
            if type(self.y[i]) != np.float64:
                self.y[i] = np.nan
        self.x = self.x.astype(np.float64)
        self.y = self.y.astype(np.float64)
        self.t = self.t.astype(np.float64)
        if verbose:
            print("Etho ML Track Converted To Numeric")

    def etho_ml_subsample(self, sample_freq: int, time_bin_size: int, verbose: bool):
        """Sub-samples the coordinates to the desired density"""
        assert self.track_type == "Ethovision Through MATLAB"
        self.x = multi_tracker.subsample(self.x, sample_freq, time_bin_size)
        self.y = multi_tracker.subsample(self.y, sample_freq, time_bin_size)
        self.t = multi_tracker.subsample(self.t, sample_freq, time_bin_size)
        if verbose:
            print("Etho ML Track Subsampled")

    def etho_ml_fill_missing(self, verbose: bool):
        """Fills in the missing coordinates"""
        assert self.track_type == "Ethovision Through MATLAB"
        self.x = multi_tracker.fill_missing_data(self.x, self.t)
        self.y = multi_tracker.fill_missing_data(self.y, self.t)
        if verbose:
            print("Etho ML Track Missing Values Filled")

    def etho_ml_convert_to_center(
        self, track_center: tuple[float, float], verbose: bool
    ):
        """Centers the coordinates based on the arena center point"""
        assert self.track_type == "Ethovision Through MATLAB"
        self.x = self.x - track_center[0]  # x coordinate of the track's center point
        self.y = self.y - track_center[1]  # x coordinate of the track's center point
        if verbose:
            print("Etho ML Track Centered")

    def anymaze_center_numeric(self, verbose: bool):
        """Changes datatype to numeric"""
        assert self.track_type == "AnyMaze Center"
        self.x = pd.to_numeric(self.x, errors="coerce")
        self.y = pd.to_numeric(self.y, errors="coerce")
        if verbose:
            print("Anymaze Center Point Track Converted To Numeric")

    def anymaze_center_running_line(
        self, running_window_length: int, window_step_size: int, verbose: bool
    ):
        """Smooths the coordinates using the running line formula"""
        assert self.track_type == "AnyMaze Center"
        # smooth the coordinates using the same function used in ethovision tracks
        self.x = multi_tracker.running_line(
            self.x, running_window_length, window_step_size
        )
        self.y = multi_tracker.running_line(
            self.y, running_window_length, window_step_size
        )
        if verbose:
            print("AnyMaze Center Point Track Smoothed")

    def anymaze_center_subsample(
        self, sample_freq: int, time_bin_size: int, verbose: bool
    ):
        """Sub-samples the coordinates to the desired density"""
        assert self.track_type == "AnyMaze Center"
        # subsample the coordinates to the desired sampling frequency
        self.x = multi_tracker.subsample(self.x, sample_freq, time_bin_size)
        self.y = multi_tracker.subsample(self.y, sample_freq, time_bin_size)
        self.t = multi_tracker.subsample(self.t, sample_freq, time_bin_size)
        if verbose:
            print("AnyMaze Center Point Track Subsampled")

    def anymaze_center_fill_missing(self, verbose: bool):
        """Fills in the missing coordinates"""
        assert self.track_type == "AnyMaze Center"
        # fill missing data using linear extrapolation
        self.x = multi_tracker.fill_missing_data(self.x, self.t)
        self.y = multi_tracker.fill_missing_data(self.y, self.t)
        if verbose:
            print("AnyMaze Center Point Track Missing Values Filled")

    def anymaze_center_convert_to_center(self, track_center, verbose):
        """Centers the coordinates based on the arena center point"""
        assert self.track_type == "AnyMaze Center"
        # these are all in pixels
        self.x = self.x - track_center[0]
        self.y = self.y - track_center[1]
        if verbose:
            print("AnyMaze Center Point Units Centered")

    def anymaze_center_convert_units(self, arena_radius_cm, trim):
        """Converts the units of the track"""
        assert self.track_type == "AnyMaze Center"
        # find the radius in pixels
        arena_radius_px_x = (np.nanmax(self.x[trim:]) - np.nanmin(self.x[trim:])) / 2
        arena_radius_px_y = (np.nanmax(self.y[trim:]) - np.nanmin(self.y[trim:])) / 2
        arena_radius_px = max(arena_radius_px_x, arena_radius_px_y)
        # convert units from pixels to cm
        self.x = self.x * (arena_radius_cm / arena_radius_px)
        self.y = self.y * (arena_radius_cm / arena_radius_px)

    def anymaze_head_numeric(self, verbose: bool):
        """Changes datatype to numeric"""
        assert self.track_type == "AnyMaze Head"
        self.x = pd.to_numeric(self.x, errors="coerce")
        self.y = pd.to_numeric(self.y, errors="coerce")
        if verbose:
            print("Anymaze Head Point Track Converted To Numeric")

    def anymaze_head_running_line(
        self, running_window_length: int, window_step_size: int, verbose: bool
    ):
        """Smooths the coordinates using the running line formula"""
        assert self.track_type == "AnyMaze Head"
        # smooth the coordinates using the same function used in ethovision tracks
        self.x = multi_tracker.running_line(
            self.x, running_window_length, window_step_size
        )
        self.y = multi_tracker.running_line(
            self.y, running_window_length, window_step_size
        )
        if verbose:
            print("AnyMaze Head Point Track Smoothed")

    def anymaze_head_subsample(
        self, sample_freq: int, time_bin_size: int, verbose: bool
    ):
        """Sub-samples the coordinates to the desired density"""
        assert self.track_type == "AnyMaze Head"
        # subsample the coordinates to the desired sampling frequency
        self.x = multi_tracker.subsample(self.x, sample_freq, time_bin_size)
        self.y = multi_tracker.subsample(self.y, sample_freq, time_bin_size)
        self.t = multi_tracker.subsample(self.t, sample_freq, time_bin_size)
        if verbose:
            print("AnyMaze Head Point Track Subsampled")

    def anymaze_head_fill_missing(self, verbose: bool):
        """Fills in the missing coordinates"""
        assert self.track_type == "AnyMaze Head"
        # fill missing data using linear extrapolation
        self.x = multi_tracker.fill_missing_data(self.x, self.t)
        self.y = multi_tracker.fill_missing_data(self.y, self.t)
        if verbose:
            print("AnyMaze Head Point Track Missing Values Filled")

    def anymaze_head_convert_to_center(self, track_center, verbose):
        """Centers the coordinates based on the arena center point"""
        assert self.track_type == "AnyMaze Head"
        # these are all in pixels
        self.x = self.x - track_center[0]
        self.y = self.y - track_center[1]
        if verbose:
            print("AnyMaze Head Point Units Centered")

    def anymaze_head_convert_units(self, arena_radius_cm, trim):
        """Converts the units of the track"""
        assert self.track_type == "AnyMaze Head"
        # find the radius in pixels
        arena_radius_px_x = (np.nanmax(self.x[trim:]) - np.nanmin(self.x[trim:])) / 2
        arena_radius_px_y = (np.nanmax(self.y[trim:]) - np.nanmin(self.y[trim:])) / 2
        arena_radius_px = max(arena_radius_px_x, arena_radius_px_y)
        # convert units from pixels to cm
        self.x = self.x * (arena_radius_cm / arena_radius_px)
        self.y = self.y * (arena_radius_cm / arena_radius_px)



from opynfield.config.user_input import UserInput
from opynfield.readin.run_all import run_all_track_types
from opynfield.readin.summary_file import summary_file
from opynfield.config.defaults_settings import Defaults
from opynfield.config.cov_asymptote import CoverageAsymptote
from opynfield.calculate_measures.calculate_measures import tracks_to_measures
from opynfield.summarize_measures.summarize_individuals import individual_measures_to_dfs
from opynfield.summarize_measures.summarize_groups import all_group_averages
from opynfield.config.model_settings import set_up_fits
from opynfield.fit_models.fit_individual_models import (
    fit_all,
    find_fit_bounds,
    re_fit_all,
)
from opynfield.fit_models.fit_group_models import group_fit_all
from opynfield.stat_test.stat_test import format_params, format_group_params, run_tests
from copy import deepcopy
from opynfield.plotting.plot_individuals import plot_all_individuals, plot_traces
from opynfield.config.plot_settings import PlotSettings
from opynfield.plotting.plot_solo_groups import plot_all_solo_groups
from opynfield.plotting.plot_solo_groups_with_individuals import (
    plot_components_of_solo_groups,
)
from opynfield.plotting.plot_group_comparisons import plot_all_group_comparisons
import os

curr_dir = os.getcwd()
path_dir = curr_dir + '/TestRunResults2'


def run():
    """This function coordinates running a full analysis on test data"""
    # create your user config settings
    user_config = UserInput(
        {"Canton S": ["Buridian Tracker"], "Canton S 2": ["Buridian Tracker"]},
        {"Canton S": "CS1", "Canton S 2": "CS2"},
        4.2,
        30,
        1,
        1,
        0.001,
        True,
        path_dir,
    )
    user_config.prep_directory()
    # read in the data
    track_list = run_all_track_types(
        user_config.groups_and_types,
        user_config.verbose,
        user_config.arena_radius_cm,
        user_config.running_window_length,
        user_config.window_step_size,
        user_config.sample_freq,
        user_config.time_bin_size,
        user_config.trim,
    )
    # set the default parameters (or override)
    test_defaults = Defaults()
    # identify functional form for PICA and PGCA (or override)
    test_cov_asymptote = CoverageAsymptote()
    # calculate measures from track data
    standard_tracks, tracks_by_groups = tracks_to_measures(
        track_list, user_config, test_defaults, test_cov_asymptote
    )
    individual_measures_dfs = individual_measures_to_dfs(
        tracks_by_groups, test_defaults, user_config
    )
    # calculate group averages of measures
    group_averages = all_group_averages(
        individual_measures_dfs, test_defaults, user_config
    )
    # set up model fit defaults
    model_params = set_up_fits()
    # fit initial models on individual track data
    fits = fit_all(individual_measures_dfs, test_defaults, model_params)
    # change bounds based on the distribution of the parameters
    fit_upper_bounds, fit_lower_bounds, fit_initial_params = find_fit_bounds(
        fits, user_config
    )
    # refit the models on individual track data with the bounds
    bounded_fits = re_fit_all(
        individual_measures_dfs,
        test_defaults,
        model_params,
        fit_upper_bounds,
        fit_lower_bounds,
        fit_initial_params,
    )
    # fit group models with the bounds
    group_fits = group_fit_all(
        individual_measures_dfs,
        test_defaults,
        model_params,
        fit_upper_bounds,
        fit_lower_bounds,
        fit_initial_params,
    )
    # format the bounded_fits to do statistical tests
    formatted_bounded_fits = format_params(
        deepcopy(bounded_fits), test_defaults, user_config
    )
    # format the group fits to save out
    format_group_params(deepcopy(group_fits), test_defaults, user_config)
    # run the stat tests with the model fits
    run_tests(formatted_bounded_fits, test_defaults, user_config)
    # plot individuals with model fits
    plot_settings = PlotSettings(group_colors={"Canton S": "b", "Canton S 2": "r"})
    # create a track summary document
    summary_file(
        tracks_by_groups,
        test_cov_asymptote,
        test_defaults,
        model_params,
        plot_settings,
        user_config,
    )
    # plot individuals with model fits
    plot_all_individuals(
        individual_measures_dfs,
        bounded_fits,
        model_params,
        test_defaults,
        plot_settings,
        user_config,
    )
    # plot individual traces
    plot_traces(tracks_by_groups, plot_settings, user_config)
    # plot groups with model fits and error bars
    plot_all_solo_groups(
        group_averages,
        group_fits,
        model_params,
        test_defaults,
        plot_settings,
        user_config,
    )
    # plot individual makeup of groups with individual and group models and error bars
    plot_components_of_solo_groups(
        individual_measures_dfs,
        bounded_fits,
        group_averages,
        group_fits,
        model_params,
        test_defaults,
        plot_settings,
        user_config,
    )
    # plot group comparisons with models and error bars
    plot_all_group_comparisons(
        group_averages,
        group_fits,
        model_params,
        test_defaults,
        plot_settings,
        user_config,
    )
    # TODO: other csv input format
    # TODO: testing code
    # TODO: other summary info in stats folder separate file
    # TODO: check the assumptions of time plots and models based on the time binning
    # TODO: add verbose setting
    return



import warnings
import numpy as np
from opynfield.config.defaults_settings import Defaults
from opynfield.config.user_input import UserInput
import pandas as pd
import os


def time_average(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    defaults: Defaults,
    user_inputs: UserInput,
) -> dict[str, dict[str, pd.DataFrame]]:
    # for all the groups
    group_avg_dfs = {}
    for group in individual_measures_dfs:
        # for all the measures of that group
        print(f"Averaging Tracks From Group {group} by time")
        g_avg_dfs = {}
        for measure in individual_measures_dfs[group]:
            if measure in defaults.time_averaged_measures:
                # remove group name
                df_to_avg = individual_measures_dfs[group][measure].drop(
                    columns=["Group"]
                )
                # get measure time average
                measure_mean = pd.DataFrame(df_to_avg.mean(axis=0, skipna=True)).T
                # get measure time SEM
                measure_sem = pd.DataFrame(df_to_avg.sem(axis=0, skipna=True, ddof=1)).T
                # combine them into one result df and save to that group
                mean_sem = pd.concat([measure_mean, measure_sem])
                mean_sem.insert(0, "Which", ["Mean", "SEM"])  # noqa
                g_avg_dfs[measure] = mean_sem
        # save that group to all the groups
        group_avg_dfs[group] = g_avg_dfs
    # we have a dict with a dict for each group with each measure's avg df
    if defaults.save_group_csvs:
        # save the dfs
        for group in group_avg_dfs:
            path = (
                user_inputs.result_path
                + "/Groups/"
                + user_inputs.groups_to_paths[group]
                + "/AverageMeasures"
            )
            os.makedirs(path, exist_ok=True)
            for df_key in group_avg_dfs[group]:
                df_path = path + "/TimeAverage_" + df_key + ".csv"
                group_avg_dfs[group][df_key].to_csv(path_or_buf=df_path, index=False)
    if defaults.save_all_group_csvs:
        # add group names to group dfs, then concat the dfs from each group together to make one
        combined_avg_dfs = {}
        # for each measure
        for measure in group_avg_dfs[group]:  # noqa
            # pull that measure from each group
            m_list = []
            for g in group_avg_dfs:
                ga_df = group_avg_dfs[g][measure]
                ga_df.insert(0, "Group", g)
                m_list.append(ga_df)
            # combine them into one df
            combined_avg_dfs[measure] = pd.concat(m_list)
        path = user_inputs.result_path + "/Groups/CombinedGroups/AverageMeasures"
        os.makedirs(path, exist_ok=True)
        for df_key in combined_avg_dfs:
            df_path = path + "/CombinedGroups_TimeAverage_" + df_key + ".csv"
            combined_avg_dfs[df_key].to_csv(path_or_buf=df_path, index=False)
            # save csv for the combined group
    return group_avg_dfs


def pair_to_cov_measure(
    group_measures_dict: dict[str, pd.DataFrame],
    response_measures: list[str],
    mode: str,
) -> tuple[pd.DataFrame, list[str]]:
    # set up a df with possible coverage values and empty lists
    measures = response_measures[:]
    measures.insert(0, mode)
    pre_sorted_values = pd.DataFrame()
    for measure in measures:
        pre_sorted_values[measure] = (
            group_measures_dict[measure].drop(columns="Group").values.flatten(order="C")
        )
    return pre_sorted_values, measures


def sort_by_cov_measure(
    possible_coverage_values: np.ndarray,
    group_measures_dict: dict[str, pd.DataFrame],
    response_measures: list[str],
    mode: str,
) -> pd.DataFrame:
    # pair up measures and the coverage they occurred at
    pre_sorted_values, measures = pair_to_cov_measure(
        group_measures_dict, response_measures, mode
    )
    # set up a df with possible coverage values and empty lists
    sorted_values = pd.DataFrame()
    for measure in measures:
        if measure == mode:
            sorted_values[measure] = possible_coverage_values
        else:
            sorted_values[measure] = sorted_values.apply(lambda x: [], axis=1)
    # then iterate through pre_sorted_values rows and append measures to lists
    for ind in pre_sorted_values.index:
        c_index = np.argmin(
            np.abs(pre_sorted_values[mode][ind] - possible_coverage_values)
        )
        for m in response_measures:
            sorted_values[m][c_index].append(pre_sorted_values[m][ind])
    return sorted_values


def n_points_average_cov_measure(
    sorted_df: pd.DataFrame, n_points: int, mode: str
) -> pd.DataFrame:
    response_columns = [item for item in sorted_df.columns if item != mode]
    exploded_df = sorted_df.explode(response_columns)
    dummy_col = "grouping"
    exploded_df[dummy_col] = np.repeat(
        range(int(1 + len(exploded_df) / n_points)), n_points
    )[: len(exploded_df)]
    for_average = exploded_df.groupby(dummy_col).agg(
        {c: list for c in response_columns} | {mode: list}
    )
    ready_for_average = for_average.reset_index().drop(columns=dummy_col)
    return ready_for_average


def average_points(ready_for_average: pd.DataFrame) -> pd.DataFrame:
    coverage_averages = pd.DataFrame()
    for col in ready_for_average:
        col_avg = col + " mean"
        col_sem = col + " sem"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            coverage_averages[col_avg] = ready_for_average[col].apply(
                lambda x: np.nanmean(x)
            )
            std = ready_for_average[col].apply(lambda x: np.nanstd(x))
            count = ready_for_average[col].apply(
                lambda x: np.count_nonzero(~np.isnan(x))
            )
            coverage_averages[col_sem] = std / np.sqrt(count)
    return coverage_averages


def make_measures_by_cov_measure(
    group_individual_measures_dfs: dict[str, pd.DataFrame],
    defaults: Defaults,
    mode: str,
) -> pd.DataFrame:
    # pull the coverage values from all the individuals out
    coverages = group_individual_measures_dfs[mode]
    # find the maximum coverage value reached by any fly in the group
    max_coverage = (
        coverages.drop(columns=["Group"]).max(axis=1, skipna=True).max(skipna=True)
    )
    # create a list (rounded to 5 digits) of all the possible coverage values
    if mode == "coverage":
        possible_coverage_values = np.round(
            np.linspace(0, max_coverage, int(max_coverage * 360 / defaults.node_size)),
            5,
        )
    else:
        possible_coverage_values = np.round(np.linspace(0, max_coverage, 2000), 5)
    # now we want to sort measure values into lists at each matching coverage value
    sorted_df = sort_by_cov_measure(
        possible_coverage_values,
        group_individual_measures_dfs,
        defaults.coverage_averaged_measures,
        mode,
    )
    # now we want to take every n points to average
    if mode == "coverage":
        n_points = defaults.n_points_coverage
    elif mode == "pica":
        n_points = defaults.n_points_pica
    else:
        n_points = defaults.n_points_pgca
    ready_for_average = n_points_average_cov_measure(sorted_df, n_points, mode)
    # then actually average the n points
    coverage_averages = average_points(ready_for_average)
    return coverage_averages


def cov_measure_average(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    defaults: Defaults,
    user_inputs: UserInput,
    mode: str,
) -> dict[str, pd.DataFrame]:
    group_measures_by_coverage = {}
    for group in individual_measures_dfs:
        print(f"Averaging Tracks From Group {group} By {mode}")
        group_measures_by_coverage[group] = make_measures_by_cov_measure(
            individual_measures_dfs[group], defaults, mode
        )
    # have a dict of group name -> df with all measures
    if defaults.save_group_csvs:
        df_group_dict = {}
        cov_avg = mode + " mean"
        cov_sem = mode + " sem"
        for group in group_measures_by_coverage:
            group_dict = {}
            path = (
                user_inputs.result_path
                + "/Groups/"
                + user_inputs.groups_to_paths[group]
                + "/AverageMeasures"
            )
            os.makedirs(path, exist_ok=True)
            for measure in defaults.coverage_averaged_measures:
                m_avg = measure + " mean"
                m_sem = measure + " sem"
                measure_df = group_measures_by_coverage[group][
                    [cov_avg, cov_sem, m_avg, m_sem]
                ].T
                df_path = path + "/" + mode + "Average_" + measure + ".csv"
                measure_df.to_csv(path_or_buf=df_path)
                group_dict[measure] = measure_df
            df_group_dict[group] = group_dict
        if defaults.save_all_group_csvs:
            # can only do the save_all_group_csvs if save_group_csvs is done
            combined_avg_dfs = {}
            for measure in defaults.coverage_averaged_measures:
                m_list = []
                for group in df_group_dict:
                    # add group column to the df
                    group_m_df = df_group_dict[group][measure]
                    group_m_df.insert(0, "Group", group)
                    m_list.append(group_m_df)
                # combine them into one df
                combined_avg_dfs[measure] = pd.concat(m_list)
            path = user_inputs.result_path + "/Groups/CombinedGroups/AverageMeasures"
            os.makedirs(path, exist_ok=True)
            for df_key in combined_avg_dfs:
                df_path = (
                    path + "/CombinedGroups_" + mode + "Average_" + df_key + ".csv"
                )
                # save csv for combined group
                combined_avg_dfs[df_key].to_csv(path_or_buf=df_path)  # , index=False)
    return group_measures_by_coverage


def n_bins_average_pcov(exploded_df: pd.DataFrame) -> pd.DataFrame:
    row_avg_lists = []
    n_avgs = int(exploded_df["bin"].iloc[-1])
    for i in range(n_avgs):
        range_df = exploded_df[exploded_df["bin"] == i]
        new_columns = [[col + " mean", col + " sem"] for col in exploded_df.columns]
        new_columns = [item for sublist in new_columns for item in sublist]
        row_avg = pd.DataFrame(
            [np.ones(shape=(len(new_columns),)) * np.nan], columns=new_columns
        )
        for col in range_df:
            if col != "bin":
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    row_avg[col + " mean"] = np.nanmean(range_df[col].astype("float"))
                    try:
                        std = np.nanstd(range_df[col])
                    except ZeroDivisionError:
                        std = np.nan
                    count = np.count_nonzero(~np.isnan(range_df[col].astype("float")))
                    try:
                        row_avg[col + " sem"] = std / np.sqrt(count)
                    except ZeroDivisionError:
                        row_avg[col + " sem"] = np.nan
        row_avg_lists.append(row_avg)
    average_df = pd.concat(row_avg_lists)
    average_df = average_df.drop(["bin mean", "bin sem"], axis=1)
    return average_df


def make_measures_by_pcov(
    group_individual_measures_dfs: dict[str, pd.DataFrame], defaults: Defaults
) -> pd.DataFrame:
    # create a list (rounded to 5 digits) of all the possible coverage values
    possible_pcov_values = np.round(np.linspace(0, 1, 1001), 5)
    # now we want to sort measure values into lists at each matching coverage value
    sorted_df = sort_by_cov_measure(
        possible_pcov_values,
        group_individual_measures_dfs,
        defaults.coverage_averaged_measures,
        "percent_coverage",
    )
    n_rows = sorted_df.shape[0]
    # give the sorted_df a number for its group of pcov values
    sorted_df["bin"] = np.repeat(
        np.arange(n_rows / defaults.n_bins_percent_coverage),
        defaults.n_bins_percent_coverage,
    )[:n_rows]
    # then we group every n bins and average all points in those bins
    response_columns = [
        item
        for item in sorted_df.columns
        if item != "percent_coverage" and item != "bin"
    ]
    exploded_df = sorted_df.explode(response_columns)
    # now take values of exploded_df within a bin group
    # and actually average the n points
    coverage_averages = n_bins_average_pcov(exploded_df)
    return coverage_averages


def percent_coverage_average(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    defaults: Defaults,
    user_inputs: UserInput,
) -> dict[str, pd.DataFrame]:
    group_measures_by_pcov = {}
    for group in individual_measures_dfs:
        print(f"Averaging Tracks From Group {group} by percent coverage")
        group_measures_by_pcov[group] = make_measures_by_pcov(
            individual_measures_dfs[group], defaults
        )
    # have a dict of group name -> df with all measures
    if defaults.save_group_csvs:
        df_group_dict = {}
        cov_avg = "percent_coverage mean"
        cov_sem = "percent_coverage sem"
        for group in group_measures_by_pcov:
            group_dict = {}
            path = (
                user_inputs.result_path
                + "/Groups/"
                + user_inputs.groups_to_paths[group]
                + "/AverageMeasures"
            )
            os.makedirs(path, exist_ok=True)
            for measure in defaults.coverage_averaged_measures:
                m_avg = measure + " mean"
                m_sem = measure + " sem"
                measure_df = group_measures_by_pcov[group][
                    [cov_avg, cov_sem, m_avg, m_sem]
                ].T
                df_path = path + "/percent_coverageAverage_" + measure + ".csv"
                measure_df.to_csv(path_or_buf=df_path)
                group_dict[measure] = measure_df
            df_group_dict[group] = group_dict
        if defaults.save_all_group_csvs:
            # can only do the save_all_group_csvs if save_group_csvs is done
            combined_avg_dfs = {}
            for measure in defaults.coverage_averaged_measures:
                m_list = []
                for group in df_group_dict:
                    # add group column to the df
                    group_m_df = df_group_dict[group][measure]
                    group_m_df.insert(0, "Group", group)
                    m_list.append(group_m_df)
                    # combine them into one df
                    combined_avg_dfs[measure] = pd.concat(m_list)
                path = (
                    user_inputs.result_path + "/Groups/CombinedGroups/AverageMeasures"
                )
                os.makedirs(path, exist_ok=True)
                for df_key in combined_avg_dfs:
                    df_path = (
                        path
                        + "/CombinedGroups_percent_coverageAverage_"
                        + df_key
                        + ".csv"
                    )
                    # save csv for combined group
                    combined_avg_dfs[df_key].to_csv(path_or_buf=df_path)
    return group_measures_by_pcov


def all_group_averages(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    test_defaults: Defaults,
    user_config: UserInput,
) -> dict[str, dict]:
    group_averages = {
        "time": time_average(individual_measures_dfs, test_defaults, user_config),
        "coverage": cov_measure_average(
            individual_measures_dfs, test_defaults, user_config, "coverage"
        ),
        "pica": cov_measure_average(
            individual_measures_dfs, test_defaults, user_config, "pica"
        ),
        "pgca": cov_measure_average(
            individual_measures_dfs, test_defaults, user_config, "pgca"
        ),
        "percent_coverage": percent_coverage_average(
            individual_measures_dfs, test_defaults, user_config
        ),
    }
    # for time its x -> group -> y
    # for cmeasures its x -> group -> df
    return group_averages



import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import os
from opynfield.config.model_settings import ModelSpecification
from opynfield.config.defaults_settings import Defaults
from opynfield.config.plot_settings import PlotSettings
from opynfield.config.user_input import UserInput


def generate_view_title(
    path: str,
    x_measure: str,
    y_measure: str,
    individual_models: bool,
    group_model: bool,
    group_error: bool,
    extension: str,
):
    """This function generates the title to save a plot as from the component parts

    :param path: the path to the result plot folder
    :type path: str
    :param x_measure: the x-measure of the plot
    :type x_measure: str
    :param y_measure: the y-measure of the plot
    :type y_measure: str
    :param individual_models: whether the models of the individuals that make up the group are plotted
    :type individual_models: bool
    :param group_model: whether the model of the group average is plotted
    :type group_model: bool
    :param group_error: whether the error of the group average is plotted
    :type group_error: bool
    :param extension: the file extension / format to save the plot in
    :type extension: str
    :return: the path to save the plot at
    :rtype: str
    """
    path = path + f"{x_measure}_vs_{y_measure}"
    parts = list()
    if individual_models:
        parts.append("_with_individual_models")
    if group_model:
        parts.append("_with_group_model")
    if group_error:
        parts.append("_with_group_error")
    path = path + "_and".join(parts)
    path = path + extension
    return path


def plot_group_individual_comparison_cmeasure(
    group: str,
    x_measure: str,
    y_measure: str,
    individuals_x: pd.DataFrame,
    individuals_y: pd.DataFrame,
    individuals_params: pd.DataFrame,
    group_ys_df: pd.DataFrame,
    group_params: pd.DataFrame,
    model_spec: ModelSpecification,
    plot_settings: PlotSettings,
    user_inputs: UserInput,
):
    """This function plots a group's coverage-measure vs y-measure average and all the component individual
    coverage-measure vs y-measure relationships

    :param group: the group name
    :type group: str
    :param x_measure: the x-measure in the plot
    :type x_measure: str
    :param y_measure: the y-measure in the plot
    :type y_measure: str
    :param individuals_x: the coverage data from the component individuals
    :type individuals_x: pd.DataFrame
    :param individuals_y: the y-measure data from the component individuals
    :type individuals_y: pd.DataFrame
    :param individuals_params: the parameters for the individual model fits
    :type individuals_params: pd.DataFrame
    :param group_ys_df: the group average (and error) data
    :type group_ys_df: pd.DataFrame
    :param group_params: the group average model fit parameters
    :type group_params: pd.DataFrame
    :param model_spec: the model settings to use
    :type model_spec: ModelSpecification
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_inputs: the user inputs to use
    :type user_inputs: UserInput
    """
    # create figure and axes objects for the plot
    fig, ax = plt.subplots()

    # first plot all the individuals
    for i in range(individuals_y.shape[0]):
        y_plot = individuals_y.iloc[i][1:].values
        x_plot = individuals_x.iloc[i][1:].values
        ax.scatter(
            x_plot,
            y_plot,
            s=plot_settings.marker_size,
            c=plot_settings.marker_color,
            alpha=0.1,
        )
        # plot their models if needed
        if plot_settings.individual_model_fit:
            params = individuals_params.iloc[i].values
            if ~np.isnan(params[0]):
                y_fit = model_spec.model.model_function(x_plot.astype(float), *params)
                ax.plot(x_plot, y_fit, c=plot_settings.marker_color, alpha=0.3)

    # plot the group error bars if needed (do first so that the points go on top)
    y_plot_group = group_ys_df[f"{y_measure} mean"].values
    y_err_group = group_ys_df[f"{y_measure} sem"].values
    x_plot_group = group_ys_df[f"{x_measure} mean"].values
    x_err_group = group_ys_df[f"{x_measure} sem"].values
    if plot_settings.group_error_bars:
        ax.errorbar(
            x_plot_group,
            y_plot_group,
            yerr=y_err_group,
            xerr=x_err_group,
            fmt="none",
            ecolor="k",
            errorevery=plot_settings.n_between_error,
            elinewidth=plot_settings.error_width,
            alpha=plot_settings.alpha,
        )
    # then plot the group averages
    ax.scatter(x_plot_group, y_plot_group, s=plot_settings.marker_size, c="k")
    # then plot their model if needed
    if plot_settings.group_model_fit:
        y_fit_group = model_spec.model.model_function(
            x_plot_group, *group_params.values
        )
        ax.plot(x_plot_group, y_fit_group, c="k", alpha=0.5)

    # set the axes labels
    ax.set_xlabel(f"{x_measure}")
    ax.set_ylabel(f"{y_measure}")
    # add the plot title
    fig.suptitle(f"{y_measure} by {x_measure} group {group}")
    # set axis limits
    if y_measure != "activity":
        # ax.set_xlim()
        ax.set_ylim((-0.1, 1.1))
    # show the figure
    if plot_settings.display_solo_group_figures:
        fig.show()
    # save the figure
    if plot_settings.save_combined_view_figures:
        path = (
            user_inputs.result_path
            + "/Groups/"
            + user_inputs.groups_to_paths[group]
            + "/Plots/individual_view_by_"
            + x_measure
            + "/"
        )
        os.makedirs(path, exist_ok=True)
        fig_path = generate_view_title(
            path,
            x_measure,
            y_measure,
            plot_settings.individual_model_fit,
            plot_settings.group_model_fit,
            plot_settings.group_error_bars,
            plot_settings.fig_extension,
        )
        fig.savefig(fname=fig_path, bbox_inches="tight")
    plt.close()
    return


def plot_group_individual_comparison_time(
    group: str,
    x_measure: str,
    y_measure: str,
    individuals_y: pd.DataFrame,
    individuals_params: pd.DataFrame,
    group_y: pd.DataFrame,
    group_params: pd.DataFrame,
    model_spec: ModelSpecification,
    plot_settings: PlotSettings,
    user_inputs: UserInput,
):
    """This function plots a group's coverage-measure vs y-measure average and all the component individual
    coverage-measure vs y-measure relationships

    :param group: the group name
    :type group: str
    :param x_measure: the x-measure in the plot
    :type x_measure: str
    :param y_measure: the y-measure in the plot
    :type y_measure: str
    :param individuals_y: the y-measure data from the component individuals
    :type individuals_y: pd.DataFrame
    :param individuals_params: the parameters for the individual model fits
    :type individuals_params: pd.DataFrame
    :param group_y: the group average (and error) data
    :type group_y: pd.DataFrame
    :param group_params: the group average model fit parameters
    :type group_params: pd.DataFrame
    :param model_spec: the model settings to use
    :type model_spec: ModelSpecification
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_inputs: the user inputs to use
    :type user_inputs: UserInput
    """
    # create figure and axes objects for the plot
    fig, ax = plt.subplots()

    # first plot all the individuals
    for i in range(individuals_y.shape[0]):
        y_plot = individuals_y.iloc[i][1:].values
        x_plot = np.arange(len(y_plot))
        ax.scatter(
            x_plot,
            y_plot,
            s=plot_settings.marker_size,
            c=plot_settings.marker_color,
            alpha=0.1,
        )
        # plot their models if needed
        if plot_settings.individual_model_fit:
            params = individuals_params.iloc[i]
            y_fit = model_spec.model.model_function(x_plot, *params)
            ax.plot(x_plot, y_fit, c=plot_settings.marker_color, alpha=0.3)

    # plot the group error bars if needed (do first so that the points go on top)
    y_plot_group = group_y.iloc[0][2:]
    y_err_group = group_y.iloc[1][2:]
    x_plot_group = np.arange(len(y_plot_group))
    if plot_settings.group_error_bars:
        ax.errorbar(
            x_plot_group,
            y_plot_group,
            yerr=y_err_group,
            xerr=None,
            fmt="none",
            ecolor="k",
            errorevery=plot_settings.n_between_error,
            elinewidth=plot_settings.error_width,
            alpha=plot_settings.alpha,
        )
    # then plot the group averages
    ax.scatter(x_plot_group, y_plot_group, s=plot_settings.marker_size, c="k")
    # then plot their model if needed
    if plot_settings.group_model_fit:
        y_fit_group = model_spec.model.model_function(
            x_plot_group, *group_params.values
        )
        ax.plot(x_plot_group, y_fit_group, c="k", alpha=0.5)

    # set the axes labels
    ax.set_xlabel("time (s)")
    ax.set_ylabel(f"{y_measure}")
    # add the plot title
    fig.suptitle(f"{y_measure} by {x_measure} group {group}")
    # set axis limits
    if y_measure not in ["activity", "percent_coverage", "pica", "pgca", "coverage"]:
        # ax.set_xlim()
        ax.set_ylim((-0.1, 1.1))
    # show the figure
    if plot_settings.display_solo_group_figures:
        fig.show()
    # save the figure
    if plot_settings.save_combined_view_figures:
        path = (
            user_inputs.result_path
            + "/Groups/"
            + user_inputs.groups_to_paths[group]
            + "/Plots/individual_view_by_"
            + x_measure
            + "/"
        )
        os.makedirs(path, exist_ok=True)
        fig_path = generate_view_title(
            path,
            x_measure,
            y_measure,
            plot_settings.individual_model_fit,
            plot_settings.group_model_fit,
            plot_settings.group_error_bars,
            plot_settings.fig_extension,
        )
        fig.savefig(fname=fig_path, bbox_inches="tight")
    plt.close()
    return


def plot_components_of_solo_groups(
    individuals: dict[str, dict[str, pd.DataFrame]],
    individual_fits: dict[str, dict[str, dict[str, pd.DataFrame]]],
    groups: dict[str, dict],
    group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]],
    model_specs: dict[str, dict[str, ModelSpecification]],
    defaults: Defaults,
    plot_settings: PlotSettings,
    user_inputs: UserInput,
):
    """This function coordinates the plotting of all the groups x-measure vs y-measure averages with the component
    individuals x-measure vs y-measure relationships

    :param individuals: the component individual y-measure data, indexed by group and y-measure
    :type individuals: dict[str, dict[str, pd.DataFrame]]
    :param individual_fits: the component individual model fit parameters, indexed by group, x-measure, and y-measure
    :type individual_fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param groups: the group average data, indexed by group (and y-measure for time)
    :type groups:  dict[str, dict]
    :param group_fits: the group average model fits, indexed by group, x-measure, and y-measure
    :type group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param model_specs: the model settings to use, indexed by x-measure and y-measure
    :type model_specs: dict[str, dict[str, ModelSpecification]]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_inputs: the user inputs to use
    :type user_inputs: UserInput
    """
    # groups is x_measure -> group -> y_measure -> averages and sems for time
    # groups is x_measure -> group -> averages and sems by name for cmeasure
    # individuals is group -> measure -> df with rows are individuals and columns are 'time' points
    # group_fits is group -> x_measure -> y_measure -> model params
    # individual_fits is group -> x_measure -> y_measure -> df with rows are individuals and cols are model params
    # model specs are x_measure -> y_measure -> ModelSpecification
    for group in groups["time"]:
        print(f"Plotting Group And Individuals of {group} by time")
        for y_measure in groups["time"][group]:
            if y_measure != "r":
                # for that group, x, and y
                # plot group average against all the individuals in that group
                plot_group_individual_comparison_time(
                    group,
                    "time",
                    y_measure,
                    individuals[group][y_measure],
                    individual_fits[group]["time"][y_measure],
                    groups["time"][group][y_measure],
                    group_fits[group]["time"][y_measure],
                    model_specs["time"][y_measure],
                    plot_settings,
                    user_inputs,
                )
        for x_measure in ["coverage", "pica", "pgca", "percent_coverage"]:
            print(f"Plotting Group And Individuals of {group} by {x_measure}")
            for y_measure in defaults.coverage_averaged_measures:
                plot_group_individual_comparison_cmeasure(
                    group,
                    x_measure,
                    y_measure,
                    individuals[group][x_measure],
                    individuals[group][y_measure],
                    individual_fits[group][x_measure][y_measure],
                    groups[x_measure][group],
                    group_fits[group][x_measure][y_measure],
                    model_specs[x_measure][y_measure],
                    plot_settings,
                    user_inputs,
                )
    return



import numpy as np
from dataclasses import dataclass
from typing import Callable


def exponential(x, a, b, c):
    """ Fits an exponential model to the data

    :param x: independent data
    :type x: np.ndarray
    :param a: scaling parameter
    :type a: float
    :param b: decay parameter
    :type b: float
    :param c: constant parameter
    :type c: float
    :return: dependent data
    :rtype: np.ndarray
    """
    # exponential decay
    y = a * np.exp(b * x) + c
    # c is asymptote and positive
    return y


def fixed_exponential(x, a, b):
    """Fits a 'fixed exponential' model to the data which anchors the curve at (0, 0)

    :param x: independent data
    :type x: np.ndarray
    :param a: scaling parameter
    :type a: float
    :param b: growth parameter
    :type b: float
    :return: dependent data
    :rtype: np.ndarray
    """
    # growth towards an asymptote
    y = a * (np.exp(b * x) - 1)
    # parameter 'a' is asymptote and negative
    return y


def linear(x, a, b):
    """Fits a linear model to the data

    :param x: independent data
    :type x: np.ndarray
    :param a: slope
    :type a: float
    :param b: intercept
    :type b: float
    :return: dependent data
    :rtype: np.ndarray
    """
    # linear
    y = a * x + b
    # no asymptote -> ignore pica and pgca results
    return y


@dataclass
class CoverageAsymptote:
    """ This dataclass associates relevant information for the model fits needed to calculate a coverage asymptote
    (for PICA and PGCA)

    Attributes:
        f_name (Callable): the function to use in the model fit - be a function that approaches an asymptote as x approaches infinity, defaults to :func:`fixed_exponential`
        asymptote_param (int): which parameter indicates the asymptote magnitude, defaults to 0
        asymptote_sign (int): is the asymptote parameter positive or negative, defaults to -1 (negative)
        initial_parameters (tuple[float]): what p0 to use in curve fitting to find the asymptote, defaults to (-0.1, -0.1)
        parameter_bounds (tuple[list[int]]): what bounds to use in curve fitting to find the asymptote, defaults to ([-10, -10], [0, 0])
        max_f_eval (int): max number of iterations the curve fitting process can take to find the asymptote, defaults to 4000
    """
    f_name: Callable = (
        fixed_exponential  # function to model time coverage relationship with
    )
    asymptote_param: int = (
        0  # which parameter of the function is the asymptote magnitude (index at 0)
    )
    asymptote_sign: int = np.sign(
        -1
    )  # what sign is the asymptote calculated with (based on bounds & initial params)
    initial_parameters: tuple[float] = (
        -0.01,
        -0.01
    )  # parameters to start model with
    parameter_bounds: tuple[list[int], list[int]] = (
        [-10, -10],
        [0, 0]
    )  # bounds on the parameters
    max_f_eval: int = 4000  # increase if you run into runtime errors



import numpy as np
import numba as nb


def path_coverage(cov_bins: np.ndarray, num_bins: float) -> np.ndarray:
    """This function formats inputs and calls `_path_coverage` to calculate the animal's coverage over time.

    :param cov_bins: which bin the animal is located in at each tracking point
    :type cov_bins: np.ndarray
    :param num_bins: the total number of bins the arena edge region is divided into
    :type num_bins: float
    :return: the coverage of the animal at each tracking point
    :rtype: np.ndarray
    """
    # convert to int to run the actual coverage function
    bins = cov_bins.astype("int")
    m = int(num_bins)
    return _path_coverage(bins, m)


@nb.jit(fastmath=True, nopython=True)
def _path_coverage(fbl: np.ndarray, num_bins: int) -> np.ndarray:
    """This function orchestrates the calculation of an animal's coverage at each tracking point. It first identifies
    the initial bin, then at each time point updates how many times each bin has been visited. From the visits at each
    bin the coverage is calculated at each time point.

    :param fbl: fly bin location - which bin the animal is located in at each tracking point
    :type fbl: np.ndarray
    :param num_bins: the total number of bins the arena edge region is divided into
    :type num_bins: int
    :return: the coverage of the animal at each tracking point
    :rtype: np.ndarray
    """
    initial_visits = np.zeros(num_bins, dtype=np.int16)
    # list of visits by bin (start at all 0)
    first_visits = update_first_visit(initial_visits, fbl[0])
    # add 1 to the bin the fly starts in
    # if fly is not at edge do nothing
    prev_visits = first_visits
    t_end = len(fbl)
    cov = np.zeros(t_end)
    # coverage at the first time point will either be 0 or 1/num_bins
    cov[0] = fly_coverage(first_visits, num_bins)
    for t in range(1, t_end):
        # skips the first point because taken care of above
        # fbl[t-1] is previous bin
        # fbl[t] is new bin
        # prev_visits is the un-updated list of number of visits to each bin
        # current_visits is the updated list of number of visits to each bin
        current_visits = update_visits(fbl[t - 1], fbl[t], prev_visits, num_bins)
        # if fly is still in edge, add one to the bin the fly visits next, plus all the bins traversed to get there
        # if fly leaves edge do nothing
        # if fly enters the edge add one to current bin

        # calculates the coverage at that time point
        cov[t] = fly_coverage(current_visits, num_bins)
        prev_visits = current_visits
    return cov


@nb.jit(fastmath=True, nopython=True)
def update_first_visit(visits: np.ndarray, first_vis: int) -> np.ndarray:
    """Creates a bin visit array for the initial time point.

    :param visits: bin visit array initialized with 0s
    :type visits: np.ndarray
    :param first_vis: bin the animal was in for the first tracking point
    :type first_vis: int
    :return: bin visit array updated with the first bin
    :rtype: np.ndarray
    """
    if first_vis != -1:
        # -1 if not in the edge region
        # fist_vis is the bin at the first time point
        visits[first_vis] += 1
    else:
        pass
    return visits


@nb.jit(fastmath=True, nopython=True)
def fly_coverage(visits: np.ndarray, m: int) -> np.ndarray:
    """Calculates the coverage of a tracking point from a bin visit array and assigns the coverage value to the coverage
        array at that index.

    :param visits: bin visit array at a tracking point
    :type visits: np.ndarray
    :param m: number of total bins
    :type m: int
    :return: an array of coverage at each tracking point, updated up to the current tracking point
    :rtype: np.ndarray
    """
    # number of visits to all bins
    # noinspection PyArgumentList
    min_visits = visits.min()
    # fraction of bins that have been visited more than that
    frac_more_than_min = (visits > min_visits).sum() / m
    v_new = min_visits + frac_more_than_min
    return v_new


@nb.jit(fastmath=True, nopython=True)
def update_visits(
    past_fly_bin: int, current_fly_bin: int, past_visits: np.ndarray, m: int
) -> np.ndarray:
    """Determines if the bin visit array needs to be updated with information from the current bin location, or if the
    bin visit counts remain unchanged. Calls :func:`update_visits_for_steps_between_bins` the bin visit array needs
    multiple updates.

    :param past_fly_bin: the bin the animal was in at the previous tracking point
    :type past_fly_bin: int
    :param current_fly_bin: the bin that the animal is in at the current tracking point
    :type current_fly_bin: int
    :param past_visits: the bin visit array at the previous tracking point
    :type past_visits: np.ndarray
    :param m: number of total bins
    :type m: int
    :return: the bin visit array at the current tracking point
    :rtype: np.ndarray
    """
    if current_fly_bin == -1 or past_fly_bin == current_fly_bin:
        # the fly left the edge, or did not move, visits did not change
        return past_visits
    elif past_fly_bin == -1:
        # the fly is entering the edge, only can assign a visit to the current bin
        past_visits[current_fly_bin] += 1
        return past_visits
    # Here we know that current_fly_bin and past_fly_bin are valid unique
    # bins between 1 and M

    # all other flies have traversed between bins
    v_new = update_visits_for_steps_between_bins(
        past_fly_bin, current_fly_bin, past_visits, m
    )
    return v_new


@nb.jit(fastmath=True, nopython=True)
def update_visits_for_steps_between_bins(
    past_fly_bin: int, current_fly_bin: int, past_visits: np.ndarray, m: int
) -> np.ndarray:
    """Updates the bin visit array in cases where the animal has traversed across multiple bins in a single step by
    identifying the direction of travel, the number of bins traversed, and the specific bins that were visited during
    the travel between bins

    :param past_fly_bin: the bin the animal was in at the previous tracking point
    :type past_fly_bin: int
    :param current_fly_bin: the bin that the animal is in at the current tracking point
    :type current_fly_bin: int
    :param past_visits: the bin visit array at the previous tracking point
    :type past_visits: np.ndarray
    :param m: number of total bins
    :type m: int
    :return: the bin visit array at the current tracking point
    :rtype: np.ndarray
    """
    # figure out what direction the fly is moving, and how many steps
    step_direction, steps_taken = sign_to_step(past_fly_bin, current_fly_bin, m)
    # step direction (-1 = clockwise, 1 = counterclockwise), steps taken = # taken

    # figure out what bins were traversed (list of bin numbers)
    bins_traveled = bins_traversed(past_fly_bin, steps_taken, step_direction, m)

    # add one to all the bins traversed
    past_visits[bins_traveled] += 1
    return past_visits


@nb.jit(fastmath=True, nopython=True)
def sign_to_step(
    past_fly_bin: int, current_fly_bin: int, num_bins: int
) -> tuple[int, int]:
    """Identifies the direction of travel and the number of bins traversed in that direction.

    :param past_fly_bin: the bin the animal was in at the previous tracking point
    :type past_fly_bin: int
    :param current_fly_bin: the bin that the animal is in at the current tracking point
    :type current_fly_bin: int
    :param num_bins: number of total bins
    :type num_bins: int
    :return: tuple of the direction the animal is traveling (1 if counterclockwise, -1 if clockwise), and the number of
        bins the animal traversed in that direction of travel
    :rtype: tuple[int, int]
    """
    # figure out which way the fly is moving
    simple_step_number = abs(past_fly_bin - current_fly_bin)
    complement_step_number = num_bins - simple_step_number
    steps_to_take = min(simple_step_number, complement_step_number)
    # assume fly walked the shorter direction between bins
    if simple_step_number == steps_to_take:
        # did not cross the modular boundary
        if current_fly_bin > past_fly_bin:
            # fly moved counterclockwise
            return 1, steps_to_take
        else:
            # fly moved clockwise
            return -1, steps_to_take
    else:
        # fly crossed the modular boundary
        if current_fly_bin > past_fly_bin:
            # fly moved clockwise
            return -1, steps_to_take
        else:
            # fly moved counterclockwise
            return 1, steps_to_take


@nb.jit(fastmath=True, nopython=True)
def bins_traversed(past_fly_bin: int, n_steps: int, step_direction: int, m) -> np.ndarray:
    """Identifies which bins were visited along a path of ``n_steps`` from the current bin in the ``step_direction``.

    :param past_fly_bin: the bin the animal was in at the previous tracking point
    :type past_fly_bin: int
    :param n_steps: the number of bins the animal traversed
    :type n_steps: int
    :param step_direction: the direction the animal is traveling (1 if counterclockwise, -1 if clockwise)
    :type step_direction: int
    :param m: number of total bins
    :type m: int
    :return: which bins were visited along the path
    :rtype: np.ndarray
    """
    start = (
        past_fly_bin + step_direction
    )  # don't count the bin that was already visited
    end = (
        past_fly_bin + (step_direction * n_steps) + step_direction
    )  # count up in the right direction
    visits_added = (
        np.arange(start, end, step_direction) % m
    )  # make it modular if it passes the 0th bin
    # visits added is a list of bin numbers
    return visits_added



from dataclasses import dataclass, field, fields

import numpy as np
import pandas as pd


@dataclass
class StandardTrack:
    """This dataclass stores all relevant original and derived measures associated with an animal's track. It is
    produced after the track standardization procedure from :func:`tracks_to_measures` is completed.

    Attributes:
        group (str): group to which the track belongs
        x (np.ndarray): the track's x coordinates
        y (np.ndarray): the track's y coordinates
        t (np.ndarray): the time the track's coordinates were recorded
        r (np.ndarray): the track's radial coordinates
        theta (np.ndarray): the track's angular coordinates
        activity (np.ndarray): the track's activity (step length between two consecutive points)
        turn (np.ndarray): the track's turn angle (the angle in degrees that the animal turned between two consecutive steps)
        p_plus_plus (np.ndarray): the raw P++ of the track. See motion probability types for more information.
        p_plus_minus (np.ndarray): the raw P+- of the track. See motion probability types for more information.
        p_plus_zero (np.ndarray): the raw P+0 of the track. See motion probability types for more information.
        p_zero_plus (np.ndarray): the raw P0+ of the track. See motion probability types for more information.
        p_zero_zero (np.ndarray): the raw P00 of the track. See motion probability types for more information.
        coverage_bins (np.ndarray): the bin the animal was in at each tracking point
        n_bins (float): the total number of bins the arena edge is divided into
        coverage (np.ndarray): the raw coverage of the track. See coverage types for more information.
        percent_coverage (np.ndarray): the percent coverage of the track. See coverage types for more information.
        pica (np.ndarray): the PICA (Percent of Individual Coverage Asymptote) of the track. See coverage types for more information.
        pica_asymptote (float): the asymptote of the animals time vs coverage model
        pgca (np.ndarray): the PGCA (Percent of Group Coverage Asymptote) of the group the track belongs to. Or an initialized dummy value of all ones. See coverage types for more information.
        pgca_asymptote (float): the asymptote of the time vs coverage model for the group the track belongs to. Or an initialized dummy value of np.nan
        p_plus_plus_given_plus (np.ndarray): the P++Given+ of the track. See motion probability types for more information.
        p_plus_minus_given_plus (np.ndarray): the P+-Given+ of the track. See motion probability types for more information.
        p_plus_zero_given_plus (np.ndarray): the P+0Given+ of the track. See motion probability types for more information.
        p_zero_plus_given_zero (np.ndarray): the P0+Given0 of the track. See motion probability types for more information.
        p_zero_zero_given_zero (np.ndarray): the P00Given0 of the track. See motion probability types for more information.
        p_plus_plus_given_any (np.ndarray): the P++GivenAny of the track. See motion probability types for more information.
        p_plus_minus_given_any (np.ndarray): the P+-GivenAny of the track. See motion probability types for more information.
        p_plus_zero_given_any (np.ndarray): the P+0GivenAny of the track. See motion probability types for more information.
        p_zero_plus_given_any (np.ndarray): the P0+GivenAny of the track. See motion probability types for more information.
        p_zero_zero_given_any (np.ndarray): the P00GivenAny of the track. See motion probability types for more information.
    """
    # dataclass for standardized track after data is read in and wrangled
    group: str  # can change to enum
    x: np.ndarray = field(repr=False)  # x coordinate
    y: np.ndarray = field(repr=False)  # y coordinate
    t: np.ndarray = field(repr=False)  # time
    r: np.ndarray = field(repr=False)  # radius
    theta: np.ndarray = field(repr=False)  # angular position
    activity: np.ndarray = field(repr=False)  # step distance
    turn: np.ndarray = field(repr=False)  # turn angle
    p_plus_plus: np.ndarray = field(repr=False)  # p++
    p_plus_minus: np.ndarray = field(repr=False)  # p+-
    p_plus_zero: np.ndarray = field(repr=False)  # p+0
    p_zero_plus: np.ndarray = field(repr=False)  # p0+
    p_zero_zero: np.ndarray = field(repr=False)  # p00
    coverage_bins: np.ndarray = field(repr=False)  # bin visited
    n_bins: float  # number of bins
    coverage: np.ndarray = field(repr=False)  # coverage
    percent_coverage: np.ndarray = field(repr=False)  # percent coverage
    pica: np.ndarray = field(repr=False)  # percent of individual coverage asymptote
    pica_asymptote: float  # individual coverage asymptote
    pgca: np.ndarray = field(repr=False)  # percent of group coverage asymptote
    pgca_asymptote: float  # individual coverage asymptote
    p_plus_plus_given_plus: np.ndarray = field(repr=False)  # p++ given +
    p_plus_minus_given_plus: np.ndarray = field(repr=False)  # p+- given +
    p_plus_zero_given_plus: np.ndarray = field(repr=False)  # p+0 given +
    p_zero_plus_given_zero: np.ndarray = field(repr=False)  # p0+ given 0
    p_zero_zero_given_zero: np.ndarray = field(repr=False)  # p00 given 0
    p_plus_plus_given_any: np.ndarray = field(repr=False)  # p++ given any
    p_plus_minus_given_any: np.ndarray = field(repr=False)  # p+- given any
    p_plus_zero_given_any: np.ndarray = field(repr=False)  # p+0 given any
    p_zero_plus_given_any: np.ndarray = field(repr=False)  # p0+ given any
    p_zero_zero_given_any: np.ndarray = field(repr=False)  # p00 given any

    @classmethod
    def to_dataframes(
        cls: "StandardTrack",
        instances: list["StandardTrack"],
        extra_fields_by_name: list[str],
    ) -> tuple[dict[str, pd.DataFrame], list[str]]:
        """This class method, called in :func:`individual_measures_to_dfs`, will create a dataframe for each dataclass
        attribute that is either an array or is included in the 'extra_fields_by_name' argument.

        :param instances: list of the standard tracks to be summarized together (usually all tracks belonging to the
            same group)
        :type instances: list[StandardTrack]
        :param extra_fields_by_name: which non-array fields also need to be stored in dataframes
        :type extra_fields_by_name: list[str]
        :return: a tuple of a dictionary of field names to dataframes and a list of all the field  names
        :rtype: tuple[dict[str, pd.DataFrame], list[str]]
        """
        # make sure we are giving it a list of standard tracks for instances
        assert all(isinstance(i, StandardTrack) for i in instances)
        # TODO: Check validity of `extra_fields_by_name`
        # get the names of all the fields of standard track
        all_fields = fields(cls)
        # since we want to save the np array ones, get list of just those that are np arrays
        array_fields_names = [f.name for f in all_fields if f.type == np.ndarray]
        # add any additional fields we want to make a df from (like pica or pgca asymptote)
        fields_to_dataframe = array_fields_names + extra_fields_by_name
        # initialize a place to store the dfs in a dict by key = name of attribute
        all_dataframes: dict[str, pd.DataFrame] = {}
        for f in fields_to_dataframe:
            # for each field we want a df from
            arrays = [getattr(i, f) for i in instances]
            # get that field from every instance and put it in a df
            all_dataframes[f] = pd.DataFrame(arrays)
            # save that df to the dict with the field name as key
        return all_dataframes, fields_to_dataframe

    def set_pgca(self, input_pgca: np.ndarray, input_pgca_a: float):
        """This is a method that re-sets the pgca and pgca_asymptote values from their dummy initialization values to
        their true values calculated during the track standardization procedure.

        :param input_pgca: the PGCA (Percent of Group Coverage Asymptote) of the group the track belongs to. See
            coverage types for more information.
        :type input_pgca: np.ndarray
        :param input_pgca_a: the asymptote of the time vs coverage model for the group the track belongs to
        :type input_pgca_a: float
        """
        self.pgca = input_pgca
        self.pgca_asymptote = input_pgca_a



import os
from collections import defaultdict
import pandas as pd
from opynfield.calculate_measures.standard_track import StandardTrack
from opynfield.config.defaults_settings import Defaults
from opynfield.config.user_input import UserInput


def individual_measures_to_dfs(
    tracks_by_groups: defaultdict[str, list[StandardTrack]],
    defaults: Defaults,
    user_inputs: UserInput,
) -> dict[str, dict[str, pd.DataFrame]]:
    group_dfs = {}
    for group in tracks_by_groups:
        print(f"Summarizing Tracks From Group {group}")
        # get a dict of df for each attribute we care about
        # save that dict to a dict by group name
        g_df, fields = StandardTrack.to_dataframes(
            tracks_by_groups[group], ["pica_asymptote", "pgca_asymptote"]
        )
        group_dfs[group] = g_df
        if defaults.save_group_csvs:
            # make a dir for results -> individuals -> group -> measures
            path = (
                user_inputs.result_path
                + "/Individuals/"
                + user_inputs.groups_to_paths[group]
                + "/Measures"
            )
            os.makedirs(path, exist_ok=True)
            for df_key in group_dfs[group]:
                df_path = path + "/" + df_key + ".csv"
                group_dfs[group][df_key].to_csv(path_or_buf=df_path, index=False)
                # save csv for the group
    if defaults.save_all_group_csvs:
        # add group names to group dfs, then concat the dfs from each group together to make one
        combined_dfs = {}
        # for each measure
        for measure in fields:  # noqa
            # pull that measure from each group
            m_list = []
            for group in group_dfs:
                g_df = group_dfs[group][measure]
                g_df.insert(0, "Group", group)
                m_list.append(g_df)
            # combine them into one df
            combined_dfs[measure] = pd.concat(m_list)
        path = user_inputs.result_path + "/Individuals/CombinedGroups/Measures"
        os.makedirs(path, exist_ok=True)
        for df_key in combined_dfs:
            df_path = path + "/CombinedGroups_" + df_key + ".csv"
            combined_dfs[df_key].to_csv(path_or_buf=df_path, index=False)
            # save csv for the combined group
    return group_dfs



from tkinter import filedialog
import pandas as pd
from opynfield.readin.track import Track
from xml.dom import minidom as md


def read_buridian(
    groups_with_file_type: list[str],
    verbose: bool,
    arena_radius_cm: float,
    running_window_length: int,
    window_step_size: int,
    sample_freq: int,
    time_bin_size: int,
    all_tracks: list[Track],
) -> list[Track]:
    """This function reads in all the tracks from the Buridian tracker format. The function extracts the x, y, and t
    information and smooths, centers, and converts the units of the track.

    :param groups_with_file_type: which groups have tracks recorded in this type
    :type groups_with_file_type: list[str]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param arena_radius_cm: the radius of the arena in which the track was recorded, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type arena_radius_cm: float
    :param running_window_length: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type running_window_length: int
    :param window_step_size: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type window_step_size: int
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param all_tracks: a list with all the Track objects from previously read-in datatypes
    :type all_tracks: list[Track]
    :return: a Track object with a consistent format for x y and t tracking points
    :rtype: list[Track]
    """
    for buridian_group in groups_with_file_type:
        if verbose:
            print(f"Running Buridian Tracker Files For Group: {buridian_group}")
        title1 = f"Select .dat files in folder for {buridian_group}"
        title2 = f"Select .xml files in folder for {buridian_group}"
        # select .dat and .xml buri files for the groups
        data_files = filedialog.askopenfilenames(
            filetypes=[("Dat files", "*.dat")], title=title1
        )
        meta_files = filedialog.askopenfilenames(
            filetypes=[("xml files", "*.xml")], title=title2
        )
        for file_num in range(len(data_files)):
            if verbose:
                print(f"{buridian_group}, File {file_num + 1} Out Of {len(data_files)}")
            # read in each file's data
            file_data = pd.read_csv(data_files[file_num], sep="\t", lineterminator="\n")
            file_data = file_data.iloc[0:-2]
            file_meta = meta_files[file_num]  # match the data and metadata
            xypx = get_meta_info(
                file_meta, arena_radius_cm
            )  # extract the metadata information
            track = Track(
                buridian_group,
                file_data["x"].values,
                file_data["y"].values,
                file_data["time"].values,
                "Buridian Tracker",
                [xypx[0], xypx[1], xypx[2]],
                False,
            )  # create the track object
            # put track coordinates through standardization procedures
            track.buri_convert_units(arena_radius_cm, xypx[2], verbose)
            track.buri_convert_to_center(xypx[0], xypx[1], verbose)
            track.buri_running_line(running_window_length, window_step_size, verbose)
            track.buri_subsample(sample_freq, time_bin_size, verbose)
            track.buri_fill_missing(verbose)
            track.standardized = True
            all_tracks.append(track)
    return all_tracks


def get_meta_info(file: str, arena_radius_cm: float) -> pd.Series:
    """This function extracts center point information from the arena in which the animal was recorded

    :param file: the path to the metadata file for this track
    :type file: str
    :param arena_radius_cm: the radius of the arena in which the track was recorded, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type arena_radius_cm: float
    :return: the x and y coordinates of the center point in cm and the radius of the arena in pixels
    :rtype: pd.Series
    """
    doc = md.parse(file)  # creates the document object from the metadata file

    # pulls the arena radius from the metadata
    rad_pix = doc.getElementsByTagName("ARENA_RADIUS")  # gets the element list
    rad_pix = rad_pix.item(0)  # gets the element
    nodes = rad_pix.childNodes  # gets the list of element nodes
    node = nodes[0]  # gets the node
    arena_radius_px = int(
        node.data
    )  # node.data is a string of the arena radius in pixels

    # pulls the x coordinate of the arena center point
    x_cen = doc.getElementsByTagName("ARENA_CENTER_X")  # gets the element list
    x_cen = x_cen.item(0)  # gets the element
    nodes = x_cen.childNodes  # gets the list of element nodes
    node = nodes[0]  # gets the node
    x_coord = int(
        node.data
    )  # node.data is a string of the pixel position of the center point x coordinate
    x_coord = x_coord * (
        arena_radius_cm / arena_radius_px
    )  # convert the point units from pixels to cm

    # pulls the y coordinate of the arena center point
    y_cen = doc.getElementsByTagName("ARENA_CENTER_Y")  # gets the element list
    y_cen = y_cen.item(0)  # gets the element
    nodes = y_cen.childNodes  # gets the list of element nodes
    node = nodes[0]  # gets the node
    y_coord = int(
        node.data
    )  # node.data is a string of the pixel position of the center point y coordinate
    y_coord = y_coord * (
        arena_radius_cm / arena_radius_px
    )  # convert the point units from pixels to cm

    # format the needed metadata into a list
    xy = pd.Series([x_coord, y_coord, arena_radius_px])
    return xy
